import math

functions = {
    'log': math.log,
    'exp': math.exp,
    'sin': math.sin,
    'cos': math.cos,
    'tan': math.tan,
    'atan': math.atan,
    'tanh': math.tanh,
    'atanh': math.atanh,
    'min': min,
    'max': max,
    'minimum': min,
    'maximum': max,
    'abs': abs
}

constants = {
    'pi': math.pi,
}
